/*
** Module   : NativeInt8.java
** Abstract : descriptor for a unsigned or signed byte 
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20140109 First version, which defaults to being an unsigned byte.
** 002 GES 20140212 Non-scalar instances must emit as a sign-extended value of the platform's
**                  word-size instead of the actual size of the instance. Also fixed the sign
**                  processing such that a negative value will report as a signed byte. This is
**                  critical to getting the proper sign-extending behavior at the native layer. 
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * A descriptor for a byte which may be either signed or unsigned. The 4GL does not differentiate
 * between the two types, but the native code may be sensitive to this.  The type will be
 * as {@link LibraryManager#TYPE_SIGNED_INT8} if the value is negative and otherwise it will
 * report as {@link LibraryManager#TYPE_UNSIGNED_INT8}.
 */
public class NativeInt8 
extends NativeInteger
{
   /**
    * Default constructor.
    */
   public NativeInt8()
   {
      super();
      
      // default to unsigned and override later if needed 
      type = LibraryManager.TYPE_UNSIGNED_INT8;
   }
   
   /**
    * Assign the data element.
    *
    * @param    value
    *           The new integer data element.
    */
   public void setValue(long value)
   {
      super.setValue(value);
      
      // override the sign
      type = (value < 0) ? LibraryManager.TYPE_SIGNED_INT8 : LibraryManager.TYPE_UNSIGNED_INT8;
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   The number of bytes needed.
    */
   public int size()
   {
      return isInArray() ? 1 : LibraryManager.ptrSize();
   }
   
   /**
    * Worker method to read the primitive value from the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected void readNative(memptr ptr)
   {
      integer num = ptr.getByte(1);
      
      if (!num.isUnknown())
      {
         value = num.longValue();
      }
   }
   
   /**
    * Worker method to write the primitive value to the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected void writeNative(memptr ptr)
   {
      if (isInArray())
      {
         ptr.setByte(value, 1);
      }
      else
      {
         writeNativeWordSize(ptr);
      }
   }
}
