/*
** Module   : library_win.c
** Abstract : implementation of operating-system specific helpers for library management
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20131024 First version.
** 002 EVL 20140127 Cycle variable needs to be declared outside the loop.
** 003 GES 20140212 Added flag to let the native layer conditionally process for loading
**                  persistent libraries. It is ignored on Windows because there is no
**                  differential behavior to implement.
*/ 

#include "library.h"
#include <windows.h>
#include <string.h>

// local function declarations
HMODULE int64ToHandle(int64_t modhandle);
int64_t handleToInt64(HMODULE handle);
int64_t pointerToInt64(FARPROC ptr);
                                          
/**
 * Attempts to load the named library.  This will use the operating-system runtime library
 * loading mechanism.
 *
 * @param    libname
 *           This may be any valid library filename for the current system, including names
 *           that contain path separators (both relative and absolute). This should not be
 *           <code>null</code> or the empty string.
 * @param    persist
 *           If <code>true</code>, the library should not be unloaded at the end of the
 *           function invocation, which means that this method should cache the module
 *           handle if it is successfully loaded.
 *
 * @return   An operating-system specific "handle" to the library. Under Windows this is
 *           the "module handle" of type HMODULE and under Linux it is just referred to as
 *           a "handle" that is returned as a void pointer.  In either case, if
 *           <code>NULL</code> (0) is returned, then the call failed.
 */ 
int64_t load_library_native(const char* libname, int persist)
{
   if (libname == NULL || strlen(libname) == 0)
      return (int64_t) 0;
   
   char* modname = strdup(libname);
   
   // out of memory, just return a failure
   if (modname == NULL)
      return (int64_t) 0;
   
   int len = strlen(modname);
   
   // on Windows (and only Windows), the 4GL implements a conversion of any forward slash
   // chars to be backslashes, for the purposes of this lookup (but not for any caching that
   // may be done of the returned handle)
   int i;
   for (i = 0; i < len; i++)
   {
      if (modname[i] == '/')
         modname[i] = '\\';
   }
   
   // at this time we have no known requirements that would force use of LoadLibraryEx
   HMODULE handle = LoadLibrary(TEXT(modname));
   
   // clean up after ourselves
   free(modname);
   
   return handleToInt64(handle);
}

/**
 * Attempts to unload the library identified by the "handle" which was returned from a
 * previous call to load_library_native.  This will use the operating-system runtime library
 * unloading mechanism. Although this call may fail, Progress seems to always silently ignore
 * any failures, so this method has no return value.
 *
 * @param    modhandle
 *           This must be a valid handle to the library.
 */ 
void unload_library_native(int64_t modhandle)
{
   FreeLibrary(int64ToHandle(modhandle));
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given symbol in the library.  The symbol name must match exactly, including the case
 * and any decorations that exist as exported in the library.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    funcname
 *           This is the exact symbol to lookup in the library.
 *
 * @return   The memory address of the function or <code>NULL</code> (0) if the function cannot
 *           be found by that name in the given library.
 */ 
int64_t find_by_name_native(int64_t modhandle, const char* funcname)
{
   FARPROC addr = GetProcAddress(int64ToHandle(modhandle), funcname);
   return pointerToInt64(addr);
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given ordinal in the library.  Only some operating systems allow exporting functions by
 * ordinal number.  Those that do allow it, will typically have a table of contiguous ordinal
 * values in the library, each of which represents a function that may or may not be also
 * exported by name (as a symbol).
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    ordinal
 *           The ordinal to lookup in the library.  Only the lower word is honored.
 *
 * @return   Linux/UNIX systems don't support ordinal exports. This always returns 0.
 */ 
int64_t find_by_ordinal_native(int64_t modhandle, uint16_t ordinal)
{
   #ifdef WORD_SIZE_64
   const char* ptr = (char*) ((int64_t) ordinal);
   #else
   const char* ptr = (char*) ((int32_t) ordinal);
   #endif
   
   // yes, windows really does "re-use" the character pointer as the ordinal, if the lower
   // word is a valid index into the ordinal table (between the starting and ending
   // ordinals, inclusive); it is ridiculous, but well, its windows
   return find_by_name_native(modhandle, ptr);
}

/**
 * Convert the given integer value into a void pointer in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 *
 * @return   The module handle as a void pointer.
 */
HMODULE int64ToHandle(int64_t modhandle)
{
   #ifdef WORD_SIZE_64
   HMODULE handle = (HMODULE) modhandle;
   #else
   HMODULE handle = (HMODULE) ((int32_t) modhandle);
   #endif
   
   return handle;
}

/**
 * Convert the given handle into an integer value in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    handle
 *           This must be a handle.
 *
 * @return   The module handle as an integer.
 */
int64_t handleToInt64(HMODULE handle)
{
   #ifdef WORD_SIZE_64
   int64_t result = (int64_t) handle;
   #else
   int32_t result = (int32_t) handle;
   #endif
   
   return (int64_t) result;
}

/**
 * Convert the given void pointer into an integer value in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    ptr
 *           This must be a pointer.
 *
 * @return   The module handle as an integer.
 */
int64_t pointerToInt64(FARPROC ptr)
{
   #ifdef WORD_SIZE_64
   int64_t result = (int64_t) ptr;
   #else
   int32_t result = (int32_t) ptr;
   #endif
   
   return (int64_t) result;
}
